#include <pthread.h>
#include <string.h>
#include <stdlib.h>
#include <semaphore.h>

#include "TaskManager.h"
#include "ChanManager.h"
#include "PeerHash.h"
#include "Network.h"
#include "Global.h"

static pthread_mutex_t chanHashLock = PTHREAD_MUTEX_INITIALIZER;
static struct ChanStruct **chanHash;

static int sendQueue[256];
static unsigned char sqrPtr = 0;
static unsigned char sqwPtr = 0;
static sem_t sqrSem, sqwSem;
static pthread_mutex_t sqLock = PTHREAD_MUTEX_INITIALIZER;


static struct ChanStruct *allocateNew(int chid, int peerid){
	struct ChanStruct *ret = malloc(sizeof(*ret));
	ret->chid = chid;
	if(peerid == -1 || peerid == nw_getNodeId())
		ret->peer = NULL;
	else
		ret->peer = ph_getPeer(peerid);
	ret->waitingInstance = NULL;
	ret->orgWriter = NULL;
	ret->next = NULL;
	pthread_mutex_init(&(ret->lock), NULL);
	
	ret->rrPtr = ret->rwPtr = ret->rbuffer = malloc(BUFFERSIZE);
	ret->ssPtr = ret->srPtr = ret->swPtr = ret->sbuffer = malloc(BUFFERSIZE);

	return ret;
}

static int min3(int i1, int i2, int i3){
	i1 = (i1 < i2 ? i1 : i2);
	return (i1 < i3 ? i1 : i3);
}

static void volatile_memcpy(volatile void *dest, volatile void *src, int n){
	int i;
	for(i = 0; i < n; ++i)
		*((unsigned char*)dest + i) = *((unsigned char*)src + i);
}

static struct ChanStruct *getChan(int chid, int peerid){
	pthread_mutex_lock(&chanHashLock);
	struct ChanStruct *ptr = chanHash[chid % NCHANBUCKETS];
	struct ChanStruct *last = NULL;
	while(ptr != NULL && ptr->chid != chid){
		last = ptr;
		ptr = ptr->next;
	}
	if(ptr == NULL){
		ptr = allocateNew(chid, peerid);
		if(last == NULL)
			chanHash[chid % NCHANBUCKETS] = ptr;
		else
			last->next = ptr;
	}
	if(ptr != NULL && ptr->peer == NULL && peerid != -1 && peerid != nw_getNodeId())
		ptr->peer = ph_getPeer(peerid);
		
	pthread_mutex_unlock(&chanHashLock);
	return ptr;
}

static void *worker(void *data){
	while(masterSwitch){
		sem_wait(&sqrSem);
		pthread_mutex_lock(&sqLock);
		int chid = sendQueue[sqrPtr++];
		pthread_mutex_unlock(&sqLock);
		sem_post(&sqwSem);
		struct ChanStruct *ptr = getChan(chid, -1);
		if(ptr == NULL){
			fprintf(stderr, ERRFIX"Error in send worker - channel not found (%s:%d)\n",__FILE__,__LINE__);
			SHUTDOWN;
		}
		
		pthread_mutex_lock(&(ptr->lock));
		int maxTransfer = (int)ptr->swPtr - (int)ptr->ssPtr;
		if(maxTransfer < 0)
			maxTransfer += BUFFERSIZE;
		if(maxTransfer > 0){
			int splitPoint = BUFFERSIZE - (int)ptr->ssPtr + (int)ptr->sbuffer;
			if(splitPoint <= maxTransfer){
				if(nw_chsend(ptr->peer, ptr->chid, ptr->ssPtr, splitPoint)){
					fprintf(stderr, ERRFIX"Error on chansend (%s:%d)\n",__FILE__,__LINE__);
					SHUTDOWN;
				}
				ptr->ssPtr = ptr->sbuffer;
				if(splitPoint < maxTransfer){
					if(nw_chsend(ptr->peer, ptr->chid, ptr->ssPtr, maxTransfer - splitPoint)){
						fprintf(stderr, ERRFIX"Error on chansend (%s:%d)\n",__FILE__,__LINE__);
						SHUTDOWN;
					}
					ptr->ssPtr += maxTransfer - splitPoint;
				}
			}
			else{
				if(nw_chsend(ptr->peer, ptr->chid, ptr->ssPtr, maxTransfer)){
					fprintf(stderr, ERRFIX"Error on chansend (%s:%d)\n",__FILE__,__LINE__);
					SHUTDOWN;
				}
				ptr->ssPtr += maxTransfer;
			}
		}
		pthread_mutex_unlock(&(ptr->lock));
	}
	return NULL;
}



int ch_init(int workers){
	chanHash = malloc(NCHANBUCKETS * sizeof(struct ChanStruct*));
	memset(chanHash, 0, NCHANBUCKETS * sizeof(struct ChanHash*));
	
	sem_init(&sqrSem, 0, 0);
	sem_init(&sqwSem, 0, 256);
	
	int i;
	for(i = 1; i <= workers; ++i){
		pthread_t t;
		if(pthread_create(&t, NULL, worker, NULL)){
			fprintf(stderr, ERRFIX"Could not create required number of threads. (Failed on %d of %d) (%s:%d)\n", i, workers, __FILE__, __LINE__);
			return -1;
		}
	}
	return 0;
}




int ch_action(struct InstanceStruct *instance){
	int chid = instance->chanTrans[instance->localCh];
	int peerid = instance->chanTrans[instance->localCh + 1];
	
	struct ChanStruct *ptr = getChan(chid, peerid);
	if(ptr == NULL){
		fprintf(stderr, ERRFIX"Could not find chan (%s:%d)\n",__FILE__,__LINE__);
		return -1;
	}
	if(ptr->peer == NULL && nw_getNodeId() != peerid){
		instance->nodeWait = peerid;
		if(instance->state == CHANR)
			instance->state = CHANRNW;
		else if(instance->state == CHANW)
			instance->state = CHANWNW;
		else{
			fprintf(stderr, ERRFIX"Erroneously state %d (%s:%d)\n", instance->state, __FILE__, __LINE__);
			return -1;
		}
		tm_requeue(instance);
		return 0;
	}
	

	pthread_mutex_lock(&(ptr->lock));
	
	unsigned char trade = 0;
	void volatile *volatile rbuffer;
	void volatile *volatile rwPtr;
	void volatile *volatile rrPtr;
	void volatile *volatile sbuffer;
	void volatile *volatile swPtr;
	void volatile *volatile srPtr;
	
	if(ptr->peer == NULL){ //Local
		if(ptr->orgWriter == NULL && instance->state == CHANW){
			ptr->orgWriter = instance;
		}
		if(ptr->orgWriter == instance){
			rbuffer = ptr->sbuffer;
			rwPtr = ptr->swPtr;
			rrPtr = ptr->srPtr;
			sbuffer = ptr->rbuffer;
			swPtr = ptr->rwPtr;
			srPtr = ptr->rrPtr;
			trade = 1;
		}
	}
	if(trade == 0){
		rbuffer = ptr->rbuffer;
		rwPtr = ptr->rwPtr;
		rrPtr = ptr->rrPtr;
		sbuffer = ptr->sbuffer;
		swPtr = ptr->swPtr;
		srPtr = ptr->srPtr;
	}

	if(instance->state == CHANR){
		int trans = 0;
		int maxTransfer = (int)rwPtr - (int)rrPtr;
		if(maxTransfer < 0)
			maxTransfer += BUFFERSIZE;
		while(instance->comSize > 0 && maxTransfer > 0){
			int splitPoint = BUFFERSIZE - (int)rrPtr + (int)rbuffer;
			int toTransfer = min3(instance->comSize, splitPoint, maxTransfer);
			volatile_memcpy(instance->comPtr, rrPtr, toTransfer);
			instance->comPtr += toTransfer;
			instance->comSize -= toTransfer;
			maxTransfer -= toTransfer;
			rrPtr += toTransfer;
			if((int)rrPtr == (int)rbuffer + BUFFERSIZE)
				rrPtr = rbuffer;
			trans += toTransfer;
		}
		if(instance->comSize == 0){
			instance->state = READY;
			tm_requeue(instance);
		}
		else{
			instance->prev = NULL;
			instance->next = ptr->waitingInstance;
			if(ptr->waitingInstance != NULL){
				ptr->waitingInstance->prev = instance;
			}
			ptr->waitingInstance = instance;
		}
		if(trans){
			if(ptr->peer == NULL){//same peer
				while(ptr->waitingInstance != NULL){
					if(ptr->waitingInstance != instance){
						struct InstanceStruct *tmp = ptr->waitingInstance;
						ptr->waitingInstance = ptr->waitingInstance->next;
						tm_requeue(tmp);
					}
					else{;
						ptr->waitingInstance->prev = NULL;
						ptr->waitingInstance = ptr->waitingInstance->next;
					}
				}
			}
			else{
				nw_chsend(ptr->peer, ptr->chid, NULL, trans);
			}
		}
	}
	else if(instance->state == CHANW){
		int trans = 0;
		int maxTransfer = (int)srPtr - (int)swPtr - 1;
		if(maxTransfer < 0)
			maxTransfer += BUFFERSIZE;
		
		while(instance->comSize > 0 && maxTransfer > 0){
			int splitPoint = BUFFERSIZE - (int)swPtr + (int)sbuffer;
			int toTransfer = min3(instance->comSize, splitPoint, maxTransfer);
			volatile_memcpy(swPtr, instance->comPtr, toTransfer);
			instance->comPtr += toTransfer;
			instance->comSize -= toTransfer;
			maxTransfer -= toTransfer;
			swPtr += toTransfer;
			trans += toTransfer;
			if((int)swPtr == (int)sbuffer + BUFFERSIZE)
				swPtr -= BUFFERSIZE;
		}
		if(instance->comSize == 0){
			instance->state = READY;
			tm_requeue(instance);
		}
		else{
			instance->prev = NULL;
			instance->next = ptr->waitingInstance;
			if(ptr->waitingInstance != NULL){
				ptr->waitingInstance->prev = instance;
			}
			ptr->waitingInstance = instance;
		}
		if(trans){
			if(ptr->peer == NULL){//same peer
				while(ptr->waitingInstance != NULL){
					if(ptr->waitingInstance != instance){
						struct InstanceStruct *tmp = ptr->waitingInstance;
						ptr->waitingInstance = ptr->waitingInstance->next;
						tm_requeue(tmp);
					}
					else{
						ptr->waitingInstance->prev = NULL;
						ptr->waitingInstance = ptr->waitingInstance->next;
					}
				}
			}
			else{
				sem_wait(&sqwSem);
				pthread_mutex_lock(&sqLock);
				sendQueue[sqwPtr++] = chid;
				pthread_mutex_unlock(&sqLock);
				sem_post(&sqrSem);
			}
		}
	}
	else{
		pthread_mutex_unlock(&(ptr->lock));
		fprintf(stderr, ERRFIX"Erroneously state %d (%s:%d)\n", instance->state, __FILE__, __LINE__);
		return -1;
	}
	if(trade){
		ptr->rbuffer = sbuffer;
		ptr->rwPtr = swPtr;
		ptr->rrPtr = srPtr;
		ptr->sbuffer = rbuffer;
		ptr->swPtr = rwPtr;
		ptr->srPtr = rrPtr;
	}
	else{
		ptr->rbuffer = rbuffer;
		ptr->rwPtr = rwPtr;
		ptr->rrPtr = rrPtr;
		ptr->sbuffer = sbuffer;
		ptr->swPtr = swPtr;
		ptr->srPtr = srPtr;
	}
	
	pthread_mutex_unlock(&(ptr->lock));
	return 0;
}


int ch_receive(int chid, void *data, int size){
	struct ChanStruct *ptr = getChan(chid, -1);
	
	if(ptr == NULL){
		fprintf(stderr, ERRFIX"Error on receive - channel not found (%s:%d)\n",__FILE__,__LINE__);
		return -1;
	}
	
	pthread_mutex_lock(&(ptr->lock));
	
	if(data == NULL){ //Ack
		int maxTransfer = (int)ptr->ssPtr - (int)ptr->srPtr;
		if(maxTransfer < 0)
			maxTransfer += BUFFERSIZE;
		if(size > maxTransfer){
			fprintf(stderr, ERRFIX"Buffer overflow %d/%d (%s:%d)\n",size,maxTransfer,__FILE__,__LINE__);
			pthread_mutex_unlock(&(ptr->lock));
			return -1;
		}
		ptr->srPtr += size;
		if(ptr->srPtr > ptr->sbuffer + BUFFERSIZE)
			ptr->srPtr -= BUFFERSIZE;
		while(ptr->waitingInstance != NULL){
				struct InstanceStruct *tmp = ptr->waitingInstance;
				ptr->waitingInstance = ptr->waitingInstance->next;
				tm_requeue(tmp);
		}
	}
	else{ //Receive
		int maxTransfer = (int)ptr->rrPtr - (int)ptr->rwPtr - 1;
		if(maxTransfer < 0)
			maxTransfer += BUFFERSIZE;
		
		if(maxTransfer < size){
			fprintf(stderr, ERRFIX"Buffer overflow %d/%d (%s:%d)\n",size,maxTransfer,__FILE__,__LINE__);
			pthread_mutex_unlock(&(ptr->lock));
			return -1;
		}
		int splitPoint = BUFFERSIZE - (int)ptr->rwPtr + (int)ptr->rbuffer;
		if(splitPoint <= size){
			volatile_memcpy(ptr->rwPtr, data, splitPoint);
			ptr->rwPtr = ptr->rbuffer;
			if(splitPoint < size){
				volatile_memcpy(ptr->rwPtr, data + splitPoint, size - splitPoint);
				ptr->rwPtr += size - splitPoint;
			}
		}
		else{
			volatile_memcpy(ptr->rwPtr, data, size);
			ptr->rwPtr += size;
		}
		while(ptr->waitingInstance != NULL){
			struct InstanceStruct *tmp = ptr->waitingInstance;
			ptr->waitingInstance = ptr->waitingInstance->next;
			tm_requeue(tmp);
		}
	}
	pthread_mutex_unlock(&(ptr->lock));
	return 0;
}