
#define _GNU_SOURCE
#include <search.h>

#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pthread.h>
#include <errno.h>

#include "FunctionManager.h"
#include "TaskManager.h"
#include "Global.h"

#define PATH "./dlibs/"
#define PERM 0777

static pthread_mutex_t tabLock = PTHREAD_MUTEX_INITIALIZER;
static struct hsearch_data *tab;

int fm_init(){
	umask(~PERM);
	tab = malloc(sizeof(*tab));
	bzero(tab, sizeof(*tab));
	
	if(hcreate_r(50, tab) == 0){
		fprintf(stderr,ERRFIX"Error creating hash table in function manager. (%s:%d)\n", __FILE__, __LINE__);
		return -1;
	}
	return 0;
}

ENTRY *fm_createFile(char *name){
	struct FunStruct *fun = malloc(sizeof(struct FunStruct));
	fun->name = malloc(strlen(name) + 1);
	strcpy(fun->name, name);
	fun->handle = NULL;
	fun->fun = NULL;
	
	char *tmp = malloc(strlen(fun->name) + strlen(PATH) + 1);
	sprintf(tmp, PATH"%s", fun->name);
	
	if((fun->file = fopen(tmp, "wb")) == NULL){
		free(tmp);
		fprintf(stderr,ERRFIX"Error creating file! (%s:%d)\n", __FILE__, __LINE__);
		return NULL;
	}
	free(tmp);
	
	ENTRY entry, *res;
	entry.key = name;
	entry.data = fun;
	int success;
	
	pthread_mutex_lock(&tabLock);
	success = hsearch_r(entry, ENTER, &res, tab);
	pthread_mutex_unlock(&tabLock);
	
	if(success == 0){
		fprintf(stderr,ERRFIX"Error creating hash entry! (%s:%d)\n", __FILE__, __LINE__);
		return NULL;
	}
	if(res->data != entry.data)
		fprintf(stderr,ERRFIX"Warning: Hash table overwrite! (%s:%d)\n", __FILE__, __LINE__); //TODO send warning to source?
	
	return res;
}

int fm_completeFile(struct FunStruct *fun){
	if(fclose(fun->file)){
		fprintf(stderr,ERRFIX"Error closing file! (%s:%d)\n", __FILE__, __LINE__);
		return -1;
	}
	fun->file = NULL;
	char *tmp = malloc(strlen(fun->name) + strlen(PATH) + 1);
	sprintf(tmp, PATH"%s", fun->name);
	fun->handle = dlopen(tmp, RTLD_LAZY);
	if (!fun->handle){
		free(tmp);
		fprintf (stderr, "Error in %s:%d - %s\n", __FILE__, __LINE__, dlerror());
		return -1;
	}
	fun->fun = dlsym(fun->handle, fun->name);
	if (!fun->fun){
		free(tmp);
		fprintf(stderr, "Error in %s:%d - %s\n", __FILE__, __LINE__, dlerror());
		return -1;
	}
	free(tmp);
	return 0;
}

int fm_writeToFile(char *name, char *data, int len, int remainder){
	ENTRY entry, *res = NULL;
	entry.key = name;
	
	pthread_mutex_lock(&tabLock);
	hsearch_r(entry, FIND, &res, tab);
	pthread_mutex_unlock(&tabLock);
	
	if(res == NULL)
		if((res = fm_createFile(name)) == NULL)
			return -1;
		
	struct FunStruct *fun = (struct FunStruct*)res->data;
	if(fun->file == NULL){
		fprintf(stderr,ERRFIX"Error in %s:%d - possible overwrite\n", __FILE__, __LINE__);
		return -1;
	}
	
	if(fwrite(data, 1, len, fun->file) < len){
		fprintf(stderr,ERRFIX"Error writing to file! (%s:%d)\n", __FILE__, __LINE__);
		return -1;
	}
	
	if(remainder == 0)
		if(fm_completeFile(fun))
			return -1;
	
	return 0;
}

int fm_loadFunction(char *name){
	struct FunStruct *fun = malloc(sizeof(struct FunStruct));
	fun->name = malloc(strlen(name) + 1);
	strcpy(fun->name, name);
	fun->handle = NULL;
	fun->fun = NULL;
	fun->file = NULL;
	
	
	char *tmp = malloc(strlen(fun->name) + strlen(PATH) + 1);
	sprintf(tmp, PATH"%s", fun->name);
	
	fun->handle = dlopen(tmp, RTLD_LAZY);
	if (!fun->handle){
		free(tmp);
		fprintf (stderr, "Error in %s:%d - %s\n", __FILE__, __LINE__, dlerror());
		return -1;
	}
	fun->fun = dlsym(fun->handle, fun->name);
	if (fun->fun == NULL){
		free(tmp);
		fprintf(stderr, "Error in %s:%d - %s\n", __FILE__, __LINE__, dlerror());
		return -1;
	}
	free(tmp);
	
	ENTRY entry, *res;
	entry.key = fun->name;
	entry.data = fun;
	int success;
	
	pthread_mutex_lock(&tabLock);
	success = hsearch_r(entry, ENTER, &res, tab);
	pthread_mutex_unlock(&tabLock);
	
	if(success == 0){
		fprintf(stderr,ERRFIX"Error creating hash entry! (%s:%d)\n", __FILE__, __LINE__);
		return -1;
	}
	if(res->data != entry.data)
		fprintf(stderr,ERRFIX"Warning: Hash table overwrite! (%s:%d)\n", __FILE__, __LINE__); //TODO send warning to source?
	
	return 0;
}

int fm_readFunction(char *name, void **ptr, unsigned long *fileLen){
	
	char *tmp = malloc(strlen(name) + strlen(PATH) + 1);
	sprintf(tmp, PATH"%s", name);
	
	FILE *file;
	file = fopen(tmp, "rb");
	if (!file){
		fprintf(stderr, "Unable to open file %s - %s - (%s:%d)\n", name, strerror(errno), __FILE__, __LINE__);
		return -1;
	}
	fseek(file, 0, SEEK_END);
	*fileLen = ftell(file);
	fseek(file, 0, SEEK_SET);
	*ptr = malloc(*fileLen);
	if(ptr == NULL){
		fprintf(stderr,ERRFIX"Unable to allocate %lu bytes of memory (%s:%d)\n", *fileLen, __FILE__, __LINE__);
		return -1;
	}
	if(fread(*ptr, 1, *fileLen, file) != *fileLen){
		fprintf(stderr,ERRFIX"Error while reading from file %s (%s:%d)\n", name, __FILE__, __LINE__);
		return -1;
	}
	fclose(file);
	free(tmp);
	return 0;
}

int fm_createInstance(char *name, int *chanTrans){
	ENTRY entry, *res = NULL;
	entry.key = name;
	
	pthread_mutex_lock(&tabLock);
	int success = hsearch_r(entry, FIND, &res, tab);
	pthread_mutex_unlock(&tabLock);
	
	if(res == NULL || success == 0){
			if(fm_loadFunction(name)){
				fprintf(stderr,ERRFIX"File not found! %s:%d\n", __FILE__, __LINE__);
				return -1;
			}
			pthread_mutex_lock(&tabLock);
			success = hsearch_r(entry, FIND, &res, tab);
			pthread_mutex_unlock(&tabLock);	
	}	
	if(res == NULL || success == 0){//Should not happen
		fprintf(stderr,ERRFIX"Entry not found! %s:%d\n", __FILE__, __LINE__);
		return -1;
	}
	
	if(tm_createNew((struct FunStruct*)res->data, chanTrans)){
		fprintf(stderr,ERRFIX"Error on creating new instance! %s:%d\n", __FILE__, __LINE__);
		return -1;
	}
	
	return 0;
	
}

