#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <pthread.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <semaphore.h>

#include "Network.h"
#include "FunctionManager.h"
#include "Global.h"
#include "PeerHash.h"
#include "ChanManager.h"

#define BLOCKS      256
#define BLOCKSIZE  1024
#define TBUFFERSIZE 2048

struct ThreadList{
	pthread_t *t;
	int listenSocket;
	struct ThreadList *next;
};


enum Mode{IDLE, PARTLY, BUFFER, TRANSFER, EXECUTE, WRITE, READ, HANDSHAKE, ERROR}; //BUFFER is unused
struct ParseRet{
	enum Mode mode;
	char *name;
	int id; //Chan id, peer id or file socket/ptr, subject to change
	int aux; //Aux data (port, meassage id)
	int n;	//number of bytes or args
};
//List of threads to cancel in panic-mode
//Must always be consistent, and be read without allocating lock.
//(Aquire lock for writing)
static pthread_mutex_t threadsLock = PTHREAD_MUTEX_INITIALIZER;
static struct ThreadList *threadsFirst = NULL;
static struct ThreadList *threadsLast = NULL;

//List of peers
static pthread_mutex_t peersLock = PTHREAD_MUTEX_INITIALIZER;
static struct PeerList *peers = NULL;
struct PeerString{
	int allocSize;
	int usedSize;
	char *string;
}peerString;


//Node info
static int nodeId = -1;
static int nodePort;

//Internal prototypes
void *receive_thread(void *);
void *recvWrk_thread(void *);


//List aux functions
static void addThread(pthread_t *t, int listenSocket){
	struct ThreadList *tmp = malloc(sizeof(*tmp));
	tmp->t = t;
	tmp->listenSocket = listenSocket;
	tmp->next = NULL;
	pthread_mutex_lock(&threadsLock);
	if(threadsFirst == NULL)
		threadsFirst = tmp;
	if(threadsLast != NULL)
		threadsLast->next = tmp;
	threadsLast = tmp;
	pthread_mutex_unlock(&threadsLock);
}

struct PeerList *addPeer(int nodeSocket, struct sockaddr_storage *client, char cascade, char isAccept){
	struct PeerList *peer = malloc(sizeof(*peer));
	peer->prev = NULL;
	peer->socket = nodeSocket;
	peer->saddr = client;
	peer->id = -1;
	peer->cascade = cascade;
	peer->rcvBuffer = malloc(BLOCKS * BLOCKSIZE);
	peer->rPtr = peer->wPtr = 0;
	sem_init(&peer->rSem, 0, 0);
	sem_init(&peer->wSem, 0, BLOCKS);
	pthread_mutex_init(&peer->sendLock, NULL);

	pthread_mutex_lock(&peersLock);

	if(isAccept){
		pthread_mutex_lock(&peer->sendLock);
		char *tmp = peerString.string;
		int left = peerString.usedSize;
		while(left){
			int n = send(peer->socket, (void*)tmp, left, 0);
			if(n < 1){
				fprintf(stderr,ERRFIX"Error on send! (%s:%d)\n", __FILE__, __LINE__);
				pthread_mutex_unlock(&peer->sendLock);
				pthread_mutex_unlock(&peersLock);
				return NULL;
			}
			left -= n;
			tmp += n;
		}
	pthread_mutex_unlock(&peer->sendLock);
	}

	peer->next = peers;
	if(peers != NULL)
		peers->prev = peer;
	peers = peer;

	char host[256];
	char port[10];
	if(getnameinfo((struct sockaddr*)client, sizeof(*client), host, 255, port, 10, NI_NUMERICHOST | NI_NUMERICSERV) == 0){
		if(isAccept)
			printf(PREFIX"[%s]:%s connected\n", host, port);
		else
			printf(PREFIX"Connected to [%s]:%s\n", host, port);
	}
	else{
		if(isAccept)
			printf(PREFIX"Node connected, IP not found (ERROR) - (%s:%d)\n", __FILE__, __LINE__);
		else
			printf(PREFIX"Connected to node, IP not found (ERROR) - (%s:%d)\n", __FILE__, __LINE__);
		return NULL;
	}
	pthread_mutex_unlock(&peersLock);

	pthread_t *t1 = malloc(sizeof (*t1));
	pthread_t *t2 = malloc(sizeof (*t2));
	pthread_create(t1, NULL, receive_thread, peer);
	pthread_create(t2, NULL, recvWrk_thread, peer);
	return peer;
}

int connectPeer(char *peerInfo){
	char *savePtr = NULL;
	char *ip = strtok_r(peerInfo, "|", &savePtr);
	char *port = strtok_r(NULL, "|", &savePtr);

	struct addrinfo hints;
	memset(&hints, 0, sizeof hints);
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE;

	struct addrinfo *list;
	if(getaddrinfo(ip, port, &hints, &list)){
		fprintf(stderr,ERRFIX"Error on getaddrinfo (%s:%d)\n", __FILE__, __LINE__);
		return -1;
	};
	struct addrinfo *ptr;
	struct PeerList *peer;
	for(ptr = list; ptr != NULL; ptr = ptr->ai_next){
		int serverSocket = socket(ptr->ai_family, ptr->ai_socktype, ptr->ai_protocol);
		if(serverSocket < 0) continue;
		if(connect(serverSocket, ptr->ai_addr, ptr->ai_addrlen) < 0) continue;

		struct sockaddr *addr = malloc(sizeof *addr);
		socklen_t len = sizeof *addr;
		if(getsockname(serverSocket, addr, &len)) continue;

		if((peer = addPeer(serverSocket, (struct sockaddr_storage*)addr, 0, 0)) == NULL) continue;
		break;
	}
	freeaddrinfo(list);
	if(peer == NULL){
		perror("Connect");
		fprintf(stderr,ERRFIX"Unable to connect (%s:%d)\n", __FILE__, __LINE__);
		return -1;
	}
	return 0;
}

int cascadePeer(char *buffer){
	char cmpl = 0;
	if(buffer[strlen(buffer) - 1] == ' ')
		cmpl = 1;

	char *savePtr = NULL;
	char *tok = strtok_r(buffer," ",&savePtr);
	char *last = NULL;
	while(tok != NULL){
		if(last != NULL)
			if(connectPeer(last))
				return -1;
		last = tok;
		tok = strtok_r(NULL," ",&savePtr);

	}
	if(cmpl){
		if(last != NULL)
			if(connectPeer(last))
				return -1;
		buffer[0] = 0;
	}
	else{
		memmove(buffer, last, strlen(last) + 1);
	}
	return 0;
}

struct ParseRet parse(char *msg){
	struct ParseRet ret = {IDLE, NULL, -1, 0};
	char *ptr;
	switch(*msg){
		case 'H': //H<id> <port> <size of list of nodes>
			ret.mode = HANDSHAKE;
			ret.id = strtol(msg + 1, &ptr, 10);
			ret.aux = strtol(ptr + 1, &ptr, 10);
			ret.n = strtol(ptr + 1, NULL, 10);
			break;
		case 'T': //T<name> <size of file>
			ret.mode = TRANSFER;
			ret.name = strndup(msg + 1, (int)strchr(msg + 1, ' ') - (int)msg - 1);
			ret.n = strtol(msg + strlen(ret.name) + 1, NULL, 10);
			break;
		case 'E': //E<name> <size of arguments>
			ret.mode = EXECUTE;
			ret.name = strndup(msg + 1, (int)strchr(msg + 1, ' ') - (int)msg - 1);
			ret.n = strtol(msg + strlen(ret.name) + 1, NULL, 10);
			break;
		case 'W': //W<chan> <size>
			ret.mode = WRITE;
			ret.id = strtol(msg + 1, &ptr, 10);
			ret.n = strtol(ptr + 1, NULL, 10);
			break;
		case 'R': //R<chan> <size>
			ret.mode = READ;
			ret.id = strtol(msg + 1, &ptr, 10);
			ret.n = strtol(ptr + 1, NULL, 10);
			break;
		default:
			ret.mode = ERROR;
	}
	return ret;
}

//Thread functions
void *receive_thread(void *data){
	struct PeerList *peer = (struct PeerList*)data;
	while(masterSwitch){
		int n = 0;
		sem_wait(&peer->wSem);
		if((n = recv(peer->socket, &peer->rcvBuffer[peer->wPtr * BLOCKSIZE + 4], BLOCKSIZE - 4, 0)) < 1)
			SHUTDOWN;
		*(short*)&peer->rcvBuffer[peer->wPtr * BLOCKSIZE] = (short)n;
		*(short*)&peer->rcvBuffer[peer->wPtr * BLOCKSIZE + 2] = 4;
		++peer->wPtr;
		sem_post(&peer->rSem);
	}
	return NULL;
}

void *recvWrk_thread(void *data){
	struct PeerList *peer = (struct PeerList*)data;

	char *buffer = malloc(TBUFFERSIZE);
	buffer[0] = 0;
	void *chanTrans;
	int ctPtr = 0;
	struct ParseRet state = {IDLE, NULL, -1, 0};

	//Wait for first item
	sem_wait(&peer->rSem);
	
	char advance = 0;
	while(masterSwitch){

		short len = *(short*)&peer->rcvBuffer[peer->rPtr * BLOCKSIZE];
		short off = *(short*)&peer->rcvBuffer[peer->rPtr * BLOCKSIZE + 2];
		char *msg = &peer->rcvBuffer[peer->rPtr * BLOCKSIZE + off];
		int partLen = strnlen(msg, len);
		char end = 0;
		if(partLen < len){
			++partLen;
			end = 1;
		}

		switch(state.mode){
			case IDLE:
				if(!end){
					strncpy(buffer, msg, len);
					state.mode = PARTLY;
				}
				else
					state = parse(msg);
				if(partLen == len)
					advance = 1;
				else{
					advance = 0;
					*(short*)&peer->rcvBuffer[peer->rPtr * BLOCKSIZE]     -= partLen; //new len
					*(short*)&peer->rcvBuffer[peer->rPtr * BLOCKSIZE + 2] += partLen; //new offset
				}
				break;
			case PARTLY:
				strncat(buffer, msg, partLen);
				if(end){
					state = parse(buffer);
					buffer[0] = 0;
				}
				if(partLen == len)
					advance = 1;
				else{
					advance = 0;
					*(short*)&peer->rcvBuffer[peer->rPtr * BLOCKSIZE]     -= partLen; //new len
					*(short*)&peer->rcvBuffer[peer->rPtr * BLOCKSIZE + 2] += partLen; //new offset
				}
				break;
			case HANDSHAKE:
				partLen = (len < state.n ? len : state.n);
				state.n -= partLen;

				if(peer->id != -1 && peer->cascade){ //Cascade on 2nd handshake
					strncat(buffer, msg, partLen);
					if(cascadePeer(buffer)){
						fprintf(stderr,ERRFIX"Could not cascade! (%s:%d)\n",__FILE__, __LINE__);
						SHUTDOWN;
					}
				}

				if(state.n == 0){
					if(peer->id == -1){
						peer->id = state.id;
						ph_add(peer);
						pthread_mutex_lock(&peersLock);
						pthread_mutex_lock(&peer->sendLock);
						char *tmp = peerString.string;
						int left = peerString.usedSize;
						while(left){
							int n = send(peer->socket, (void*)tmp, left, 0);
							if(n < 1){
								fprintf(stderr,ERRFIX"Error on send! (%s:%d)\n", __FILE__, __LINE__);
								pthread_mutex_unlock(&peer->sendLock);
								pthread_mutex_unlock(&peersLock);
								SHUTDOWN;
							}
							left -= n;
							tmp += n;
						}

						char host[256];
						char port[10];
						if(getnameinfo((struct sockaddr*)peer->saddr, sizeof(*peer->saddr), host, 255, port, 10, NI_NUMERICHOST | NI_NUMERICSERV) == 0){
							printf(PREFIX"Handshake from [%s]:%s - id:%d\n", host, port, state.id);
							if(peerString.allocSize < peerString.usedSize + 22){
								peerString.allocSize *= 2;
								char *nstring = malloc(peerString.allocSize);
								strcpy(nstring, peerString.string);
								free(peerString.string);
								peerString.string = nstring;
							}
							sprintf(peerString.string + peerString.usedSize - 1, "%s|%d ", host, state.aux);
							peerString.usedSize = strlen(peerString.string + 17) + 18;
							sprintf(peerString.string, "H%04d %05d %04d", nodeId, nodePort, peerString.usedSize - 17);
						}
						else{
							printf(PREFIX"Client connected (ip unavailible)\n");
							return NULL;
						}
						pthread_mutex_unlock(&peer->sendLock);
						pthread_mutex_unlock(&peersLock);
					}
					state.mode = IDLE;
				}
				if(partLen == len)
					advance = 1;
				else{
					advance = 0;
					*(short*)&peer->rcvBuffer[peer->rPtr * BLOCKSIZE]     -= partLen; //new len
					*(short*)&peer->rcvBuffer[peer->rPtr * BLOCKSIZE + 2] += partLen; //new offset
				}
				break;
			case TRANSFER:
				partLen = (len < state.n ? len : state.n);
				state.n -= partLen;

				if(fm_writeToFile(state.name, msg, partLen, state.n)){
					fprintf(stderr,ERRFIX"Error writing to file! (%s:%d)\n", __FILE__, __LINE__);
					SHUTDOWN;
				}
				if(state.n == 0){
					free(state.name);
					state.mode = IDLE;
				}

				if(partLen == len)
					advance = 1;
				else{
					advance = 0;
					*(short*)&peer->rcvBuffer[peer->rPtr * BLOCKSIZE]     -= partLen; //new len
					*(short*)&peer->rcvBuffer[peer->rPtr * BLOCKSIZE + 2] += partLen; //new offset
				}
				break;
			case EXECUTE:
				partLen = (len < state.n ? len : state.n);
				state.n -= partLen;
				
				if(ctPtr == 0)
					chanTrans = malloc(state.n + partLen);
				memcpy(chanTrans + ctPtr, msg, partLen);
				ctPtr += partLen;

				if(state.n == 0)
					state.mode = IDLE;

				if(partLen == len)
					advance = 1;
				else{
					advance = 0;
					*(short*)&peer->rcvBuffer[peer->rPtr * BLOCKSIZE]     -= partLen; //new len
					*(short*)&peer->rcvBuffer[peer->rPtr * BLOCKSIZE + 2] += partLen; //new offset
				}


				if(state.mode == IDLE){
					if(fm_createInstance(state.name, chanTrans)){
						fprintf(stderr,ERRFIX"Error on execute! (%s:%d)\n", __FILE__, __LINE__);
						SHUTDOWN;
					}
					ctPtr = 0;
					free(state.name);
					state.name = NULL;
				}
				break;
			case WRITE:
				partLen = (len < state.n ? len : state.n);
				state.n -= partLen;
				if(ch_receive(state.id, msg, partLen)){
					fprintf(stderr,ERRFIX"Error on write (%s:%d)\n",__FILE__,__LINE__);
					SHUTDOWN;
				}
				if(state.n == 0)
					state.mode = IDLE;
				if(partLen == len)
					advance = 1;
				else{
					advance = 0;
					*(short*)&peer->rcvBuffer[peer->rPtr * BLOCKSIZE]     -= partLen; //new len
					*(short*)&peer->rcvBuffer[peer->rPtr * BLOCKSIZE + 2] += partLen; //new offset
				}
				break;
			case READ:
				if(ch_receive(state.id, NULL, state.n)){
					fprintf(stderr,ERRFIX"Error on read (%s:%d)\n",__FILE__,__LINE__);
					SHUTDOWN;
				}
				state.mode = IDLE;
				break;
			case ERROR:
			default:
				fprintf(stderr,ERRFIX"Error (%s:%d)\n",__FILE__,__LINE__);
				SHUTDOWN;
				break;
		}
		if(advance && state.mode != READ){
			//Advance to next queue item
			++peer->rPtr;
			sem_post(&peer->wSem);
			sem_wait(&peer->rSem);
			advance = 0;
		}
	}
	return NULL;
}

void *accept_thread(void *data){
	while(masterSwitch){
		struct sockaddr_storage *client;
		size_t size = sizeof(*client);
		client = malloc(size);
		int clientSocket = accept(*(int*)data, (struct sockaddr *)client, &size); //Cancel point
		if(clientSocket < 0)
			fprintf(stderr,ERRFIX"Error on accept\n");
		else if(addPeer(clientSocket, client, 0, 1) == NULL)
			SHUTDOWN;

	}
	return NULL;
}

//Public functions
int nw_init(int id, int port, char *mip, char *mpt){
	nodeId = id;

	peerString.string = malloc(1024);
	peerString.allocSize = 1024;
	peerString.usedSize = 19;
	sprintf(peerString.string, "H%04d %05d 0002", id, port);
	sprintf(peerString.string + 17, " ");

	struct addrinfo hints;
	memset(&hints, 0, sizeof hints);
	hints.ai_family = AF_INET; //CHANGE TO AF_NET6 for IPv6
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE;

	struct addrinfo *list;

	char portStr[6];
	sprintf(portStr, "%d", port);

	if(getaddrinfo(NULL, portStr, &hints, &list)){
		fprintf(stderr,ERRFIX"Error on getaddrinfo (%s:%d)\n", __FILE__, __LINE__);
		return -1;
	}

	int errorListen = -1;
	struct addrinfo *ptr;
	for(ptr = list; ptr != NULL; ptr = ptr->ai_next){
		int yes = 1;
		int serverSocket = socket(ptr->ai_family, ptr->ai_socktype, ptr->ai_protocol);
		if(serverSocket < 0) continue;
		if(setsockopt(serverSocket, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) < 0) continue;
		if(bind(serverSocket, ptr->ai_addr, ptr->ai_addrlen) < 0) continue;
		if(listen(serverSocket, 10) < 0) continue;

		pthread_t *t = malloc(sizeof(*t));
		int *socket = malloc(sizeof(int));
		*socket = serverSocket;

		if(pthread_create(t, NULL, accept_thread, socket)){
			free(t);
			free(socket);
			continue;
		}
		addThread(t, serverSocket);
		errorListen = 0;
	}
	freeaddrinfo(list);
	if(errorListen){
		perror("Listen");
		fprintf(stderr,ERRFIX"Unable to socket/bind/listen (%s:%d)\n", __FILE__, __LINE__);
		return -1;
	}

	if(mip != NULL && mpt != NULL){
		if(getaddrinfo(mip, mpt, &hints, &list)){
			fprintf(stderr,ERRFIX"Error on getaddrinfo (%s:%d)\n", __FILE__, __LINE__);
			return -1;
		};
		struct addrinfo *ptr;
		struct PeerList *peer;
		for(ptr = list; ptr != NULL; ptr = ptr->ai_next){
			int serverSocket = socket(ptr->ai_family, ptr->ai_socktype, ptr->ai_protocol);
			if(serverSocket < 0) continue;
			if(connect(serverSocket, ptr->ai_addr, ptr->ai_addrlen) < 0) continue;

			struct sockaddr *addr = malloc(sizeof *addr);
			socklen_t len = sizeof *addr;
			if(getsockname(serverSocket, addr, &len)) continue;

			if((peer = addPeer(serverSocket, (struct sockaddr_storage*)addr, 1, 0)) == NULL) continue;
			break;
		}
		freeaddrinfo(list);
		if(peer == NULL){
			fprintf(stderr,ERRFIX"Unable to connect (%s:%d)\n", __FILE__, __LINE__);
			return -2;
		}
	}

	return 0;

}

int nw_getNodeId(){
	return nodeId;
}

int nw_sendFile(int id, char *name){
	struct PeerList *peer = ph_getPeer(id);
	if(peer == NULL){
		fprintf(stderr,ERRFIX"Trying to transfer to non-excisting peer - id=%d name=%s (%s:%d)\n", id, name, __FILE__, __LINE__);
		return -1;
	}
	void *ptr = NULL;
	unsigned long size;
	if(fm_readFunction(name, &ptr, &size)){
		fprintf(stderr,ERRFIX"Error! (%s:%d)\n", __FILE__, __LINE__);
		return -1;
	}
	char *header = malloc(strlen(name) + 15);
	sprintf(header, "T%s %lu", name, size);
	pthread_mutex_lock(&peer->sendLock);
	int left = strlen(header) + 1;
	while(left){
		int n = send(peer->socket, (void*)header, left, 0);
		if(n < 1){
			fprintf(stderr,ERRFIX"Error on send! (%s:%d)\n", __FILE__, __LINE__);
			pthread_mutex_unlock(&peer->sendLock);
			return -1;
		}
		left -= n;
		header += n;
	}
	left = size;
	while(left){
		int n = send(peer->socket, (void*)ptr, left, 0);
		if(n < 1){
			fprintf(stderr,ERRFIX"Error on send! (%s:%d)\n", __FILE__, __LINE__);
			pthread_mutex_unlock(&peer->sendLock);
			return -1;
		}
		left -= n;
		ptr += n;
	}
	pthread_mutex_unlock(&peer->sendLock);
	return 0;
}

int nw_spawn(int id, char *name, int *chanTrans, int ctSize){
	struct PeerList *peer = ph_getPeer(id);
	if(peer == NULL){
		fprintf(stderr,ERRFIX"Trying to execute on non-excisting peer (%s:%d)\n",__FILE__,__LINE__);
		return -1;
	}
	char *buffer = malloc(strlen(name) + 12 + ctSize);
	sprintf(buffer, "E%s %d", name, ctSize);
	memcpy(buffer + strlen(buffer) + 1, chanTrans, ctSize);
	pthread_mutex_lock(&peer->sendLock);
	int left = strlen(buffer) + ctSize + 1;
	while(left){
		int n = send(peer->socket, (void*)buffer, left, 0);
		if(n < 1){
			fprintf(stderr,ERRFIX"Error on send! (%s:%d)\n", __FILE__, __LINE__);
			pthread_mutex_unlock(&peer->sendLock);
			return -1;
		}
		left -= n;
		buffer += n;
	}
	pthread_mutex_unlock(&peer->sendLock);
	return 0;
}

int nw_chsend(struct PeerList *peer, int chid, volatile void * volatile data, int size){
	char *buffer = malloc(32);
	char *tmp = buffer;
	if(data == NULL)//R<id> <size>\0
		sprintf(buffer, "R%d %d", chid, size);
	else//W<id> <size>\0<data>
		sprintf(buffer, "W%d %d", chid, size);
	pthread_mutex_lock(&peer->sendLock);
	int left = strlen(buffer) + 1;
	while(left){
		int n = send(peer->socket, (void*)buffer, left, 0);
		if(n < 1){
			fprintf(stderr,ERRFIX"Error on send! (%s:%d)\n", __FILE__, __LINE__);
			pthread_mutex_unlock(&peer->sendLock);
			return -1;
		}
		left -= n;
		buffer += n;
	}
	free(tmp);
	if(data){
		left = size;
		while(left){
			int n = send(peer->socket, (void*)data, left, 0);
			if(n < 1){
				fprintf(stderr,ERRFIX"Error on send! (%s:%d)\n", __FILE__, __LINE__);
				pthread_mutex_unlock(&peer->sendLock);
				return -1;
			}
			left -= n;
			data += n;
		}
	}
	pthread_mutex_unlock(&peer->sendLock);
	return 0;
}

int nw_panic(){
	printf(PREFIX"Panic!\n");
	pthread_mutex_lock(&peersLock);
	struct ThreadList *pt;
	for(pt = threadsFirst; pt != NULL; pt = pt->next){
		shutdown(pt->listenSocket, 2);
		pthread_cancel(*pt->t);
	}
	struct PeerList *pp;
	for(pp = peers; pp != NULL; pp = pp->next)
		shutdown(pp->socket, SHUT_RDWR);
	pthread_mutex_unlock(&peersLock);
	return 0;
}
