#include <pthread.h>
#include <stdlib.h>
#include <string.h>

#include "FunctionManager.h"
#include "TaskManager.h"
#include "PeerHash.h"
#include "Global.h"
#include "Network.h"
#include "ChanManager.h"

//Add to last, take from first;
static pthread_mutex_t queueLock = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t queueCond = PTHREAD_COND_INITIALIZER;
static struct InstanceStruct *first, *last;

static void addToQueue(struct InstanceStruct *instance){
	pthread_mutex_lock(&queueLock);
	instance->next = last;
	instance->prev = NULL;
	if(last != NULL)
		last->prev = instance;
	last = instance;
	if(first == NULL)
		first = instance;
	else if(first->prev == NULL)
		first->prev = instance;
	pthread_cond_broadcast(&queueCond);
	pthread_mutex_unlock(&queueLock);
}

//Blocks if queue is empty
static struct InstanceStruct *getFirstItem(){
	struct InstanceStruct *ret;
	pthread_mutex_lock(&queueLock);
	while(first == NULL)
		pthread_cond_wait(&queueCond, &queueLock);
	ret = first;
	if(first != NULL){
		if(first->prev){
			first->prev->next = NULL;
			first = first->prev;
		}
		else{
			first = NULL;
			last = NULL;
		}
	}
	pthread_mutex_unlock(&queueLock);
	return ret;
}

static void *worker(void *data){
	while(masterSwitch){
		struct InstanceStruct *item = getFirstItem();
		if(item->state & (READY | NEW))
			item->funStruct->fun(item);
		switch(item->state){
			case NEW:
				fprintf(stderr,ERRFIX"Task has invalid state (%s:%d)\n",__FILE__,__LINE__);
				SHUTDOWN;
				break;
			case READY:
				addToQueue(item);
				break;
			case CHANR:
			case CHANW:
				if(ch_action(item)){
					fprintf(stderr,ERRFIX"Error on ch_action (%s:%d)\n", __FILE__, __LINE__);
					SHUTDOWN;
				}
				break;
			case CHANRNW:
			case CHANWNW:
			case NODEWAIT:
				ph_waitForNode(item, item->nodeWait);
				break;
			case TRANSFER:
				if(nw_sendFile(((struct SpawnStruct*)item->comPtr)->peerId,
					((struct SpawnStruct*)item->comPtr)->name)){
						fprintf(stderr,ERRFIX"Error on transfer (%s:%d)\n",__FILE__,__LINE__);
						SHUTDOWN;
				}
				item->state = READY;
				addToQueue(item);
				break;
			case SPAWN:
				if(((struct SpawnStruct*)item->comPtr)->peerId == nw_getNodeId()){
					int *chanTrans = malloc(((struct SpawnStruct*)item->comPtr)->ctSize * sizeof(int));
					memcpy(chanTrans, ((struct SpawnStruct*)item->comPtr)->chanTrans, ((struct SpawnStruct*)item->comPtr)->ctSize * sizeof(int));
					if(fm_createInstance(((struct SpawnStruct*)item->comPtr)->name, chanTrans)){
								fprintf(stderr,ERRFIX"Error on spawn (%s:%d)\n",__FILE__,__LINE__);
								SHUTDOWN;
							  }
				}
				else if(nw_spawn(((struct SpawnStruct*)item->comPtr)->peerId,
					((struct SpawnStruct*)item->comPtr)->name,
					((struct SpawnStruct*)item->comPtr)->chanTrans,
					((struct SpawnStruct*)item->comPtr)->ctSize)){
						fprintf(stderr,ERRFIX"Error on spawn (%s:%d)\n",__FILE__,__LINE__);
						SHUTDOWN;
				}
				item->state = READY;
				addToQueue(item);;
				break;
			case DONE:
				free(item->chanTrans);
				free(item);
				break;
		}
	}
	return NULL;
}


int tm_init(int workers){
	first = NULL;
	last = NULL;
	int i;
	for(i = 1; i <= workers; ++i){
		pthread_t t;
		if(pthread_create(&t, NULL, worker, NULL)){
			fprintf(stderr,ERRFIX"Could not create required number of threads. (Failed on %d of %d) (%s:%d)\n", i, workers, __FILE__, __LINE__);
			return -1;
		}
	}
	return 0;
}

int tm_createNew(struct FunStruct *funStruct, int *chanTrans){
	struct InstanceStruct *instance = malloc(sizeof *instance);
	instance->funStruct = funStruct;
	instance->memPtr = NULL;
	instance->chanTrans = chanTrans;
	instance->step = 0;
	instance->state = NEW;
	addToQueue(instance);
	return 0;
}

int tm_requeue(struct InstanceStruct *instance){
	addToQueue(instance);
	return 0;
}


